#ifndef SOCKETSTREAM_EXAMPLE_ASYNCH_RESOLVER_H
#define SOCKETSTREAM_EXAMPLE_ASYNCH_RESOLVER_H

#include <memory>
#include <boost/thread.hpp>
#include <socketstream/resolver.h>

namespace ss
{

	/**
	 * @brief Asynchronous name resolution service.
	 *
	 * This implementation is POSIX Thread specific.
	 */
	class asynch_resolver : public resolver
	{
	public:
	
		asynch_resolver () {}

		/**
		 * @brief Asynchronously resolve names.
		 *
		 * Overrides base class method with the same signature.
		 */
		void resolve (const char* name, const char* serv)
		{
			m_thread = std::auto_ptr<boost::thread>(
					new boost::thread(
						asynch_resolve(this, name, serv)
					)
				);
		}

		/**
		 * @brief Returns result of name resolution.
		 *
		 * Blocks until name resolution finishes.
		 */
		resolver_result const& result () const
		{
			m_thread->join();
			return static_cast<resolver const*>(this)->result();
		}

	private:
	
		// Not CopyConstructible.
		asynch_resolver (asynch_resolver const&);
		
		// Not Assignable.
		asynch_resolver operator= (asynch_resolver const&);

		class asynch_resolve
		{
		public:
			asynch_resolve(asynch_resolver* that_,
			               char const* name_,
			               char const* serv_)
			: that(that_), name(name_), serv(serv_) {}

			void operator() () { that->resolve(name, serv);	}
			
			typedef void result_type;
			
		private:
			resolver*   that;
			const char* name;
			const char* serv;
		};
		
		std::auto_ptr<boost::thread> m_thread;

	};

}

#endif // SOCKETSTREAM_EXAMPLE_ASYNCH_RESOLVER_H
