// -*- Socket Streams -*-

// Copyright (C) 2003, 2004, 2005 Pedro Lamarão <pedro.lamarao@mndfck.org>

// This file is part of the Socket Streams Library.  This library is free
// software; you can redistribute it and/or modify it under the
// terms of the GNU Lesser General Public License as published by the
// Free Software Foundation; either version 2.1 of the License
// or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License along
// with this library; see the file COPYING. If not, write to the Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
// USA

/** @file basic_socket.h
 * @brief Contains the socket class.
 */

#ifndef SOCKETSTREAM_BASIC_SOCKET_H
#define SOCKETSTREAM_BASIC_SOCKET_H 1

#include <socketstream/config.h>

#include <socketstream/sockets_base.h>
#include <socketstream/socket_exception.h>

namespace ss {
  /**
   * @brief Socket class wraps up the C socket interface.
   * 
   * This class implements "resource acquisition is initialization" for
   * socket descriptors.
   *
   * The purpose of this class is to merely wrap the C socket API (effectively,
   * to wrap the socket descriptor).
   * Thus, it's member function prototypes follow closely those of the C
   * API, and no error treatment other than the trivial is done.
   *
   * The only big departure from the C API is the fact that error conditions
   * intercepted are thrown as socket_exception objects containing the
   * error code from std::errno. Thus, almost all functions now return void.
   *
   * Also, the return type of the accept method is basic_socket, instead of the
   * descriptor type.
   * 
   * All I/O operations are considered non-const operations; setsockopt
   * is also considered non-const for obvious reasons.
   * getsockopt, getpeername, getsockname, and all "properties" are
   * considered const operations.
   */
  class SOCKETSTREAM_API basic_socket
  {
  public:

    /**
     * @brief This is the type of a socket descriptor on the system.
     * 
     * This is usually int, but don't count on it unless you're Unix-only.
     */
    typedef sockets_base::socket_type descriptor_t;
    
    template <typename _CharT, typename _TraitsT>
    friend class basic_socketbuf;

    /**
     * @brief Default constructor.
     *
     * Socket's constructor wraps the socket() system call,
     * which allocates a socket descriptor in the system for this process.
     * This constructor will create a socket of default family and style,
     * currently sockets_base::inet and sockets_base::stream.
     */
    basic_socket ();

    /**
     * @brief Constructor.
     * 
     * @param __family Socket's address family.
     * @param __style Socket's communication style.
     * @param __protocol Socket's communication protocol.
     * 
     * Socket's constructor wraps the socket() system call,
     * which allocates a socket descriptor in the system for this process.
     * 
     * Constructs socket of specified family, style and protocol.
     * We usually use a default protocol.
     */
    basic_socket (int __family, int __style, int __protocol = 0);

    /**
     * @brief Constructor for an existing socket descriptor.
     * 
     * @param __sd Socket descriptor.
     * 
     * This constructor does no call socket(), and just associates itself
     * with the specified socket descriptor.
     *
     * basic_socket takes ownership of the descriptor, and will close it
     * at destruction time.
     *
     * Particularly useful with basic_socket::accept().
     */
    basic_socket (descriptor_t __sd);

    /**
     * Destructor.
     */
    ~basic_socket ();

    /**
     * @brief Bind socket to local address.
     * 
     * @param __addr Pointer to buffer containing addres to bind to.
     * @param __size Size of address buffer.
     * 
     * This will set the local address, which is
     * usually only important in preparation for a listen().
     */
    void
    bind (sockaddr const* __addr, socklen_t __size);

    /**
     * @brief Connect socket to remote address.
     * 
     * @param __addr Pointer to buffer containing address to connect to.
     * @param __size Size of address buffer.
     * 
     * This will attempt to connect at a remote (listening) socket bound to the
     * specified address.
     * This is only truly meaningful for stream sockets.
     */
    void
    connect (sockaddr const* __addr, socklen_t __size);

    /**
     * @brief Listen to incoming connections.
     * 
     * @param __backlog (can be safely left on default).
     * 
     * Puts socket in a listening state, so it can accept connections.
     * 
     * This is the opposite of connect(), so to speak, and
     * is only truly meaningful for stream sockets.
     */
    void
    listen (int __backlog = 5);

    /**
     * @brief Accept incoming connection.
     * 
     * @returns Descriptor for connected socket.
     * 
     * Accepts an incoming connection on a listening socket, and returns
     * a socket descriptor connected to that peer ready for I/O.
     */
    descriptor_t
    accept ();

    /**
     * @brief Receive data on connected socket.
     * 
     * @param __s Pointer to input buffer.
     * @param __n Size of input buffer.
     * @param __flags I/O flags.
     * 
     * @return Number of bytes received.
     */
    int
    receive (char* __s, size_t __n, int __flags = 0);

    /**
     * @brief Receive data on connectionless socket.
     * 
     * @param __s Pointer to input buffer.
     * @param __n Size of input buffer.
     * @param __addr Pointer to buffer to store sender address.
     * @param __size Size of address buffer.
     * @param __flags I/O flags.
     * 
     * @return Number of bytes received.
     * 
     * This function is only truly meaningful with connectionless sockets.
     */
    int
    receive (char* __s, size_t __n,
             sockaddr* __addr, socklen_t* __size,
             int __flags = 0);

    /**
     * @brief Send data on connected socket.
     * 
     * @param __s Pointer to buffer containing data to send.
     * @param __n Size of data buffer.
     * @param __flags I/O flags.
     * 
     * @return Number of bytes sent.
     */
    int
    send (const char* __s, size_t __n, int __flags = 0);

    /**
     * @brief Send data on connectionless socket.
     *
     * @param __s Pointer to buffer containing data to send.
     * @param __n Size of data buffer.
     * @param __addr Pointer to buffer containing address to send to.
     * @param __size Size of the address buffer.
     * @param __flags I/O flags.
     * 
     * @return Number of bytes sent.
     * 
     * This function is only truly meaningful with connectionless sockets.
     */
    int
    send (const char* __s, size_t __n,
          sockaddr const* __addr, socklen_t __size,
          int __flags = 0);

    /**
     * @brief Set socket option.
     * 
     * @param __level Option level.
     * @param __option Option code.
     * @param __value Pointer to buffer with option value.
     * @param __size Size of option value.
     * 
     * This function sets a socket option with setsockopt().
     */
    void
    set_option (int __level, int __option,
                char const* __value, socklen_t __size);

    /**
     * @brief Get socket option.
     * 
     * @param __level Option level.
     * @param __option Option code.
     * @param __value Pointer to buffer to store option value.
     * @param __size Pointer to socklen_t to store option value size.
     * 
     * This function obtains a socket option value with getsockopt().
     */
    void
    get_option (int __level, int __option,
                char* __value, socklen_t* __size) const;

    /**
     * @brief Shut the socket down.
     * @param __mode Mode.
     * 
     * This will close the connection with the peer for sending, receiving
     * or both.
     */
    void
    shutdown (int __mode = sockets_base::rw);
    
    /**
     * @brief Peer address.
     * 
     * @param __name Peer address.
     * @param __size Size of address structure.
     */
    void
    peer_address (sockaddr* __name, socklen_t* __size) const;
    
    /**
     * @brief Socket address.
     * 
     * @param __name socket address.
     * @param __size Size of adsress structure.
     */
    void
    socket_address (sockaddr* __name, socklen_t* __size) const;

    /**
     * @brief Socket descriptor.
     * 
     * @return Socket descriptor.
     */
    descriptor_t
    descriptor () const
    { return _M_sd; }    

  private:

    // Not CopyConstructible.
    basic_socket (basic_socket const&);

    // Not Assignable.
    basic_socket operator= (basic_socket const&);

    descriptor_t  _M_sd;

  };
}

#endif	// SOCKETSTREAM_BASIC_SOCKET_H
