// -*- Socket Streams -*-

// Copyright (C) 2003, 2004, 2005 Pedro Lamarão <pedro.lamarao@mndfck.org>

// This file is part of the Socket Streams Library.  This library is free
// software; you can redistribute it and/or modify it under the
// terms of the GNU Lesser General Public License as published by the
// Free Software Foundation; either version 2.1 of the License
// or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License along
// with this library; see the file COPYING. If not, write to the Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
// USA

/** @file basic_socketstream_listener.h
 * @brief Contains the listener template.
 */

#ifndef SOCKETSTREAM_BASIC_LISTENER_H
#define SOCKETSTREAM_BASIC_LISTENER_H 1

#include <socketstream/config.h>

#include <memory>

#include <socketstream/sockets_base.h>
#include <socketstream/basic_iosocketstream.h>

namespace ss {
  /** @ingroup listener
   * @brief Controls a listener socket.
   * 
   * A listener socket accepts incoming connections, returning stream objects
   * representing them.
   */
  template <typename _CharT, typename _Traits>
  class basic_listener {
  public:

    //@{
    /**
     * @brief These are standard types.
     * 
     * They permit a standardized way of referring to names of
     * (or names dependant on) the template parameters, which are specific 
     * to the implementation.
     */
    typedef _CharT     char_type;
    typedef _Traits  traits_type;
    
    typedef typename traits_type::int_type  int_type;
    typedef typename traits_type::pos_type  pos_type;
    typedef typename traits_type::off_type  off_type;
    //@}

    typedef basic_socketstream<char_type, traits_type> socketstream_type;

    /**
     * @brief Constructor.
     */
    basic_listener ()
    : _M_socket(sockets_base::def::family, sockets_base::stream)
    { }

    /**
     * @brief Constructor.
     */
    basic_listener (sockets_base::family __f)
    : _M_socket(__f, sockets_base::stream)
    { }

    /**
     * @brief Destructor.
     */
    ~basic_listener ()
    { }

    /**
     * @brief Bind socket to local address.
     * 
     * @param __addr Address to bind.
     */
    template <typename AddressT>
    void
    bind (AddressT& __addr)
    { _M_socket.bind(&__addr, sizeof(__addr)); }

    /**
     * @brief Listen for incoming connections.
     * 
     * @param __queue Size of incoming queue. Mostly ignored.
     */
    void
    listen (int __queue = 5)
    { _M_socket.listen(__queue); }

    /**
     * @brief Accept incoming connection.
     * 
     * @return Smart pointer to a connected stream.
     */
    std::auto_ptr<socketstream_type>
    accept ()
    {
      try
      {
        basic_socket::descriptor_t __s = _M_socket.accept();
        return std::auto_ptr<socketstream_type>(new socketstream_type(__s));
      }
      catch (const socket_exception& se)
      {
        switch (se.code())
        {
          case EWOULDBLOCK:
            return std::auto_ptr<socketstream_type>(0);
          case EINTR:
          default:
            throw(se);
        }
      }
    }

  protected:
    basic_socket _M_socket;
  };

  typedef basic_listener<char, std::char_traits<char> > listener;

}

#endif // SOCKETSTREAM_BASIC_LISTENER_H

