// -*- Socket Streams -*-

// Copyright (C) 2003, 2004, 2005 Pedro Lamarão <pedro.lamarao@mndfck.org>

// This file is part of the Socket Streams Library.  This library is free
// software; you can redistribute it and/or modify it under the
// terms of the GNU Lesser General Public License as published by the
// Free Software Foundation; either version 2.1 of the License
// or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License along
// with this library; see the file COPYING. If not, write to the Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
// USA

/** @file inet_factory.h
 * @brief Contains the IPv4 address class.
 */

#ifndef SOCKETSTREAM_INET_ADDRESS_H
#define SOCKETSTREAM_INET_ADDRESS_H 1

#include <socketstream/config.h>

#ifndef WIN32
# include <netinet/in.h>
#endif

namespace ss {

	/** @ingroup address
	 * @brief Improved sockaddr_in.
	 */
	class SOCKETSTREAM_API ipv4_address : protected sockaddr_in
	{
	public:

		/**
		 * @brief Base sockaddr type.
		 */
		typedef sockaddr_in sockaddr_t;

		/**
		 * @brief Address family code.
		 */
		static const int AF = AF_INET;

		/**
		 * @brief Any address.
		 */
		static const in_addr any;

		/**
		 * @brief The loopback address.
		 */
		static const in_addr loopback;

		/**
		 * @brief Default constructor.
		 */
		ipv4_address ();

		/**
		 * @brief Constructor.
		 * 
		 * Takes an IPv4 address and a transport port.
		 * 
		 * Parameters are expected in network byte order.
		 * 
		 * For a randomly chosen address, pass in ipv4_address::any
		 * as first argument.
		 * For a randomly chosen port, pass in 0 as second argument.
		 */
		ipv4_address (in_addr const& address, unsigned short port);

		/**
		 * @brief Copy constructor.
		 */
		ipv4_address (ipv4_address const& that);

		/**
		 * @brief Copy constructor.
		 */
		ipv4_address (sockaddr_in const& that);

		/**
		 * @brief Copy constructor.
		 * 
		 * Checks if the address structure is of the correct family.
		 */
		ipv4_address (sockaddr const& that);

		/**
		 * @brief Assignment operator.
		 */
		ipv4_address& operator= (ipv4_address const& that);

		/**
		 * @brief Assignment operator.
		 */
		ipv4_address& operator= (sockaddr_in const& that);

		/**
		 * @brief Assignment operator.
		 * 
		 * Checks if the address structure is of the correct family.
		 */
		ipv4_address& operator= (sockaddr const& that);

	};

}

#endif // SOCKETSTREAM_INET_ADDRESS_H
