// -*- Socket Streams -*-

// Copyright (C) 2003, 2004, 2005 Pedro Lamarão <pedro.lamarao@mndfck.org>

// This file is part of the Socket Streams Library.  This library is free
// software; you can redistribute it and/or modify it under the
// terms of the GNU Lesser General Public License as published by the
// Free Software Foundation; either version 2.1 of the License
// or (at your option) any later version.

// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License along
// with this library; see the file COPYING. If not, write to the Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
// USA

/** @file sockets_base.h
 * @brief Contains the sockets_base class.
 */

#ifndef SOCKETSTREAM_SOCKETS_BASE_H
#define SOCKETSTREAM_SOCKETS_BASE_H 1

#include <socketstream/config.h>

#ifndef WIN32
# include <sys/types.h>
# include <sys/socket.h>
#endif

/**
 * @brief Everything declared by the Socket Stream Library
 * is within namespace ss.
 */
namespace ss {

  enum __Family   { };
  enum __Style    { };
  enum __Protocol { };
  enum __Shutdown_Mode { };
  enum __Io_Flag { };
  
  inline __Io_Flag
  operator& (__Io_Flag __a, __Io_Flag __b)
  { return __Io_Flag(static_cast<int>(__a) & static_cast<int>(__b)); }
    
  inline __Io_Flag
  operator| (__Io_Flag __a, __Io_Flag __b)
  { return __Io_Flag(static_cast<int>(__a) | static_cast<int>(__b)); }

  inline __Io_Flag
  operator^ (__Io_Flag __a, __Io_Flag __b)
  { return __Io_Flag(static_cast<int>(__a) ^ static_cast<int>(__b)); }

  inline __Io_Flag
  operator&= (__Io_Flag __a, __Io_Flag __b)
  { return __a = __a & __b; }

  inline __Io_Flag
  operator|= (__Io_Flag __a, __Io_Flag __b)
  { return __a = __a | __b; }

  inline __Io_Flag
  operator^= (__Io_Flag __a, __Io_Flag __b)
  { return __a = __a ^ __b; }

  inline __Io_Flag
  operator~ (__Io_Flag __a)
  { return __Io_Flag(~static_cast<int>(__a)); }

  /**
   * @brief Base class for all socketstream specific common stuff.
   * 
   * This class is inherited by the iosocketstream classes.
   */
  class SOCKETSTREAM_API sockets_base {
  public:

    /**
     * @brief This is the socket descriptor type in the system.
     * 
     * This is usually int, but not on Windows.
     */
#ifdef WIN32
    typedef SOCKET socket_type;
    typedef SOCKET descriptor_t;
#else
    typedef int socket_type;
    typedef int descriptor_t;
#endif

    /**
     * @brief Enumerates protocol (or address) families.
     * 
     * sockets_base::family members are to be passed as the family argument to
     * basic_socket() .
     */
    typedef __Family family;

    /**
     * @brief Enumerates socket styles.
     * 
     * sockets_base::style members are to be passed as the style argument to
     * basic_socket() .
     */
    typedef __Style style;
    
    /**
     * @brief Enumerates socket protocols.
     */
    typedef __Protocol protocol;

    /**
     * @brief Enumerates shutdown modes.
     * 
     * sockets_base::shutmode members are to be passed as the mode argument to
     * basic_socket::shutdown() .
     */
    typedef __Shutdown_Mode shutmode;

    /**
     * @brief Enumerates I/O flags.
     * 
     * sockets_base::io_flags members are to be passed as the flags argument to
     * basic_socket::receive() and basic_socket::send() .
     */
    typedef __Io_Flag flag;

    /**
     * @brief Default values.
     * 
     * These should be set by the user at configure time, and used by
     * default constructors throughout the library.
     */
     struct SOCKETSTREAM_API def
    { 
      static const sockets_base::family    family;
      static const sockets_base::style     style;
      static const sockets_base::protocol  protocol;
      static const sockets_base::flag      flag;
    };

    /**
     * Destructor.
     */
    virtual
    ~sockets_base () {}

    static const family local;
    static const family ipv4;
    static const family ipv6;
    static const family netlink;
    static const family packet;

    static const style stream;
    static const style datagram;
    static const style raw;
    static const style rdm;
    static const style seqpacket;

    static const shutmode r;
    static const shutmode w;
    static const shutmode rw;

    static const flag oob;
    static const flag dontroute;
    static const flag dontwait;
    static const flag nosignal;
    static const flag peek;
    static const flag waitall;
    static const flag trunc;
    static const flag errqueue;

  };
}

#endif // SOCKETSTREAM_SOCKETS_BASE_H
